#include<stdio.h>
#include<stdlib.h>
#include "magic.h"

struct tnode {
	char *value;
	struct tnode *left;
	struct tnode *right;
};

#define MAX_WORD_LEN 30

struct tnode *root;

int insert(char *value);
struct tnode *insertHelper(struct tnode *p, char *value);
int contains(char *value);
void printTreeHelper(struct tnode *curr);
void printTree();
//struct tnode *locate(char *value);
int search(struct tnode *localRoot, char *value, struct tnode **x, struct tnode **parent, int *found);
int removeN(char *value,int n);
int delete(char *value);
int loadData(FILE *file);
int searchData(FILE *file);

int main() {
    FILE *balanced, *sorted, *random, *worst;

    balanced = fopen("words256Kbalanced.txt", "r");
    sorted   = fopen("words256Ksorted.txt", "r");
    random   = fopen("words256Krandom.txt", "r");
    worst    = fopen("words256Kworst.txt", "r");

    if (balanced == NULL ||
        sorted == NULL ||
        random == NULL ||
        worst == NULL)
    {
        fprintf(stderr, "open failed\n");
		return -1;
	}

	if (loadData(balanced)) {
        fprintf(stderr, "load data failed");
    }

	simicsBreakpoint(); /* load ruby */

    /* init state */
	searchData(sorted);
    fseek(sorted, 0, SEEK_SET); /* return to beginning of file */
    
	simicsBreakpoint(); /* start */
	if (searchData(sorted)) {
        fprintf(stderr, "search sorted failed\n");
        return -1;
    }
	simicsBreakpoint(); /* stop */

    fseek(sorted, 0, SEEK_SET); /* return to beginning of file */
    /* init state */
	searchData(sorted);
    
	simicsBreakpoint(); /* start */
	if (searchData(random)) {
        fprintf(stderr, "search random failed\n");
        return -1;
    }
    simicsBreakpoint(); /* stop */

    fseek(sorted, 0, SEEK_SET); /* return to beginning of file */
    /* init state */
	searchData(sorted);

	simicsBreakpoint(); /* start */
	if (searchData(worst)) {
        fprintf(stderr, "search worst failed\n");
        return -1;
    }
	simicsBreakpoint(); /* stop */

    close(balanced);
    close(sorted);
    close(random);
    close(worst);

	return 0;
}

void printTree() {
	printTreeHelper(root);
}

void printTreeHelper(struct tnode *curr) {
	char *left, *right;
	if(curr != NULL) {
		printf("\t\t******* %s *******\n", curr->value);
		if (curr->left != NULL) {
			left = curr->left->value;
		} else {
			left = "NULL";
		}
		if (curr->right != NULL) {
			right = curr->right->value;
		} else {
			right = "NULL";
		}
		
		printf("\t %s\t\t%s\n",left, right);
		printTreeHelper(curr->left);
		printTreeHelper(curr->right);
	}
}

	

int insert(char *valueToInsert) {
	struct tnode **rootptr;
	struct tnode *curr;
	struct tnode *next;
	int direction;
	char *value = (char *)malloc(strlen(valueToInsert)+1);
	if (value == NULL) {
		return -1;
	}
	strcpy(value, valueToInsert);

	if (root == NULL) {
		curr = (struct tnode*)malloc(sizeof(struct tnode));
		if(curr == NULL) {
			return -1;
		}
		curr->left = NULL;
		curr->right = NULL;
		curr->value = value;
		root = curr;
	} else {
		root = insertHelper(root, value);
	}
	if (root == NULL) {
		return -1;
	}
	return 0;
}

struct tnode *insertHelper(struct tnode *p, char *value) {
	int direction;
	struct tnode *node;
	if (p == NULL) {
		node = (struct tnode *)malloc(sizeof(struct tnode));
		node->value = value;
		node->left = NULL;
		node->right = NULL;
		return node;
	}
	
	direction = strcmp(value, p->value);
	if (direction < 0) {
		p->left = insertHelper(p->left, value);

	} else if (direction == 0) {
        /* already there */

	} else { /*direction > 0 */
		p->right = insertHelper(p->right, value);
	}
	return p;
}

/* returns 1 on true and 0 on false */
int contains(char *value) {
	int direction;
	struct tnode *curr;

	curr = root;
	while(curr != NULL) {
		direction = strcmp(value, curr->value);
		if (direction < 0) {
			curr = curr->left;
		} else if (direction == 0) {
			return 1;
        } else { //direction > 0 
			curr = curr->right;
		}
	}
	return 0;
}

/**
 * @param localRoot pointer to the root of the subtree to examine
 * @param value The string to look for
 * @param mainNode Used to return the address of the desired node
 * @param parent used to return the parent of the located node
 * @param used to return if the node was found
 *
 * @return 0 if value not foudn in tree with root localRoot, and 1 otherwise.
 */
int search(struct tnode *localRoot, char *value, struct tnode **mainNode, struct tnode **parent, int *found) {
	int direction;
	*found = 0;
	*parent = NULL;
	*mainNode = NULL;
	while(localRoot != NULL && !(*found)) {
		direction = strcmp(value, localRoot->value);
		if (direction < 0) {
			*parent = localRoot;
			localRoot = localRoot->left;
		} else if (direction == 0) {
			*found = 1;
			*mainNode = localRoot;
		} else {
			*parent = localRoot;
			localRoot = localRoot->right;
		}
	}
	return *found;
}

int delete(char *value) {
	return removeN(value, -1);
}

/** if n == -1 then remove all */
int removeN(char *value,int n) {
	struct tnode *parent, *nodeToRemove, *curr, *prev;
	int found;
	if (root == NULL) {
		return -1;
	}

	search(root, value, &nodeToRemove, &parent, &found);
	if (!found) {
		return -1;
	}

	if(nodeToRemove->left != NULL && nodeToRemove->right != NULL) {
		curr = nodeToRemove->right;
		prev = NULL;
		while (curr->left != NULL) {
			prev = curr;
			curr = prev->left;
		}

		if (prev != NULL) {
			prev->left = curr->left;
			curr->right = nodeToRemove->right;
		}
		
		curr->left = nodeToRemove->left;
		if (parent->left == nodeToRemove) {
			parent->left = curr;
		} else {
			parent->right = curr;
		}
		free(nodeToRemove->value);
		free(nodeToRemove);
		return(0);
	} else {
		if (nodeToRemove->left != NULL) {
			if (parent == NULL) {
				root = nodeToRemove->left;
			} else {
				curr = nodeToRemove->left;
				prev = NULL;
				while (curr->right != NULL) {
					prev = curr;
					curr = prev->right;
				}
				if (prev != NULL) {
					prev->right = curr->right;
					curr->left = nodeToRemove->left;
				}
				curr->right = nodeToRemove->right;
				if (parent->left == nodeToRemove) {
					parent->left = curr;
				} else {
					parent->right = curr;
				}
			}
		} else if (nodeToRemove->right != NULL) {
			if (parent == NULL) {
				root = nodeToRemove->right;
			} else {
				curr = nodeToRemove->right;
				prev = NULL;
				while (curr->left != NULL) {
					prev = curr;
					curr = prev->left;
				}
				if (prev != NULL) {
					prev->left = curr->left;
					curr->right = nodeToRemove->right;
				}
				curr->left = nodeToRemove->left;
				if (parent->left == nodeToRemove) {
					parent->left = curr;
				} else {
					parent->right = curr;
				}
			}
		} else {
			if (parent == NULL) {
				root = NULL;
			} else {
				if(parent->left == nodeToRemove) {
					parent->left = NULL;
				} else {
					parent->right = NULL;
				}
			}
		}
	}
	free(nodeToRemove->value);
	free(nodeToRemove);
	return 0;
}


int loadData(FILE *file) {
	char word[MAX_WORD_LEN];

	while (fscanf(file, "%s", word) > 0) {
		if (insert(word) != 0) {
			return -1;
		}
	}

	return 0;
}


int searchData(FILE *file) {
	char word[MAX_WORD_LEN];

	while (fscanf(file, "%s", word) > 0) {
		if (!contains(word)) {
			return -1;
		}
	}

	return 0;
}


